; FILE: Source:partc.ASM          REV: 5 --- Gravity Particle compo entry
; History
;    5  11th Jun 1998: updated to 12:4 fixedpoint
;

; This thing was coded by Harry "Piru" Sintonen <sintonen@jyu.fi> for
; #amycoders gravity particle compo. Modify it as you like...
;                                          signed - Piru

        machine mc68020

        incdir  includes:

        include "exec/types.i"
        include "exec/libraries.i"

        include "intuition/intuition.i"
        include "intuition/screens.i"

        include "exec/exec_lib.i"
        include "graphics/graphics_lib.i"
        include "intuition/intuition_lib.i"

call    MACRO
        jsr     (_LVO\1,a6)
        ENDM


SCR_WIDTH       EQU     256     ; better not mod:
SCR_HEIGHT      EQU     256
SCR_DEPTH       EQU     8
WINDOW_MODE     EQU     1
NUM_PARTICLES   EQU     4096

INITSPREAD      EQU     79      ; mod these if you like:
INITSXPOS       EQU     128-83
INITSYPOS       EQU     128+26
INITMAXSPEED    EQU     1<<4

mainen  move.l  sp,d0
        eor.l   d0,_seed

        bsr     openlibs
        tst.l   d0
        beq.b   .nolibs

        bsr     openscreenetc
        tst.l   d0
        beq.b   .noscreen

        ; Initialize effect stuff
        bsr     _InitEffect
        tst.l   d0
        beq.b   .cleanup

.loop
        ; Now render to chunky in _ChunkyPtr
        ; using _RenderWidth and _RenderHeight

        bsr     _Effect

        move.l  _c2p_ptr,a0             ; do c2p
        jsr     (a0)

        move.l  _GfxBase,a6
        call    WaitTOF

        IFNE    WINDOW_MODE
        move.l  _Window,a0
        move.l  (wd_UserPort,a0),a0
        move.l  (4).w,a6
        call    GetMsg
        tst.l   d0
        beq.b   .loop
        move.l  d0,a1
        call    ReplyMsg
        ELSE
        btst    #6,$BFE001
        bne.b   .loop
        ENDC

        ; Clean up effect stuff
.cleanup        bsr     _CleanupEffect

        bsr     closescreenetc
.noscreen
        bsr     closelibs

.nolibs moveq   #0,d0
        rts


        CNOP    0,4
_seed   dc.l    $BADBABE5
dornd   movem.l d7,-(sp)
        move.l  _seed,d7
        move.l  d7,d0
                lsl.l   #8,d7
        lsr.l   #8,d0
        add.l   d0,d7
        moveq   #0,d0
        move.w  d7,d0
        neg.l   d7
        divu.w  d1,d0
        swap    d0
        move.l  d7,_seed
        movem.l (sp)+,d7
        rts


        CNOP    0,4
_InitEffect
        cmp.l   #SCR_WIDTH,_RenderWidth
        bne     .bad
        cmp.l   #SCR_HEIGHT,_RenderHeight
        bne     .bad

        lea     _ParticleArray,a0
        move.l  #NUM_PARTICLES,d2
.loop
        move.w  #INITSPREAD,d1
        bsr     dornd
        IFNE    (INITSXPOS-INITSPREAD/2)
        add.l   #INITSXPOS-INITSPREAD/2,d0
        ENDC
        asl.w   #4,d0
        move.w  d0,(a0)+

        move.w  #INITSPREAD,d1
        bsr     dornd
        IFNE    (INITSYPOS-INITSPREAD/2)
        add.l   #INITSYPOS-INITSPREAD/2,d0
        ENDC
        asl.w   #4,d0
        move.w  d0,(a0)+

        move.w  #INITMAXSPEED,d1
        bsr     dornd
        sub.l   #INITMAXSPEED/2,d0
        move.w  d0,(a0)+

        move.w  #INITMAXSPEED,d1
        bsr     dornd
        sub.l   #INITMAXSPEED/2,d0
        move.w  d0,(a0)+

        subq.l  #1,d2
        bne.b   .loop

        moveq   #1,d0
        rts

.bad    moveq   #0,d0
        rts


_CleanupEffect  moveq   #1,d0
        rts


        CNOP    0,4
_Effect
        movem.l d0-a6,-(sp)

        move.l  _ChunkyPtr,a0
        move.l  #256*256,d0
        add.l   d0,a0
        lsr.l   #4,d0
.clear  clr.l   -(a0)
        clr.l   -(a0)
        clr.l   -(a0)
        clr.l   -(a0)
        subq.l  #1,d0
        bne.b   .clear

        move.l  _ChunkyPtr,a0
        lea     _ParticleArray,a1
        jsr     _gpart

        move.l  _ChunkyPtr,a0
        add.l   #128<<8|128,a0
        move.b  #2,(-256,a0)
        move.b  #2,(-1,a0)
        move.b  #2,(0,a0)
        move.b  #2,(1,a0)
        move.b  #2,(256,a0)

        movem.l (sp)+,d0-a6
        rts


openlibs
        movem.l d1-a6,-(sp)
        moveq   #0,d7
        move.l  (4).w,a6
        lea     (.GfxName,pc),a1
        moveq   #39,d0
        call    OpenLibrary
        move.l  d0,_GfxBase
        beq.b   .exit
        lea     (.IntuiName,pc),a1
        moveq   #39,d0
        call    OpenLibrary
        move.l  d0,_IntuiBase
        beq     .closegfx

        moveq   #1,d7
        bra.b   .exit
.closegfx
        move.l  _GfxBase,a1
        call    CloseLibrary
        clr.l   _GfxBase

.exit   move.l  d7,d0
        movem.l (sp)+,d1-a6
        rts

.GfxName
        dc.b    'graphics.library',0
.IntuiName
        dc.b    'intuition.library',0
        CNOP    0,2


closelibs
        movem.l d1-a6,-(sp)
        move.l  (4).w,a6

        move.l  _IntuiBase,a1
        call    CloseLibrary
        clr.l   _IntuiBase

        move.l  _GfxBase,a1
        call    CloseLibrary
        clr.l   _GfxBase

        movem.l (sp)+,d1-a6
        rts



        IFNE    WINDOW_MODE

openscreenetc
        movem.l d1-a6,-(sp)
        moveq   #0,d7

        ; First open screen with some tags
        ; Will get screen with default overscan.

        sub.l   a0,a0
        lea     (.windowtags,pc),a1
        move.l  _IntuiBase,a6
        call    OpenWindowTagList
        move.l  d0,_Window
        beq     .exit

        ; Cool, we have a window! ;)
        ; Now we need to see whether we got native or
        ; GFXCard screen and choose between c2p and
        ; WriteChunkyPixels/WritePixelArray8.
        ; Here we go:

        lea     _gfxcard_c2p_v39,a0
        move.l  (4).w,a6
        cmp.w   #40,(LIB_VERSION,a6)    ; V40 system?
        blo.b   .set_ptr                ; nopes, use WritePixelArray8
        lea     _gfxcard_c2p_v40,a0     ; yes, use WriteChunkyPixels
.set_ptr
        move.l  a0,_c2p_ptr


        move.l  _Window,a0
        moveq   #0,d0
        move.b  (wd_BorderLeft,a0),d0
        addq.l  #1,d0
        move.l  d0,_RenderXstart
        move.b  (wd_BorderTop,a0),d0
        addq.l  #1,d0
        move.l  d0,_RenderYstart

        move.l  (wd_RPort,a0),a0
        move.l  a0,_RastPort
        move.l  (rp_BitMap,a0),a0
        move.l  a0,_BitMap
        move.l  (bm_Planes,a0),_Plane0

        move.l  _BitMap,a0
        moveq   #BMA_WIDTH,d1           ; width please
        move.l  _GfxBase,a6
        call    GetBitMapAttr
        move.l  d0,_Width
        lsr.l   #3,d0
        move.l  d0,_BytesPerRow

        move.l  #SCR_WIDTH,_RenderWidth

        move.l  _BitMap,a0
        moveq   #BMA_HEIGHT,d1          ; height please
        call    GetBitMapAttr
        move.l  d0,_Height

        move.l  #SCR_HEIGHT,_RenderHeight

        move.l  _Height,d0
        move.l  _Width,d1
        mulu.w  d0,d1
        lsr.l   #3,d1
        move.l  d1,_PlaneSize           ; width*height/8

        cmp.l   #_gfxcard_c2p_v39,_c2p_ptr
        bne.b   .no_temprp

        ; temporary rastport (copy of rp with Layer set == NULL,
        ;temporary memory allocated for
        ;temprp->BitMap with Rows set == 1,
        ;temprp->BytesPerRow == (((width+15)>>4)<<1),
        ;and temporary memory allocated for
        ;temprp->BitMap->Planes[])

        move.l  _RastPort,a0
        lea     _temprp,a1
        move.l  #rp_SIZEOF,d0
        move.l  (4).w,a6
        call    CopyMem
        lea     _temprp,a1
        clr.l   (rp_Layer,a1)

        move.l  _RenderWidth,d0         ; width
        moveq   #1,d1                   ; height=1
        moveq   #SCR_DEPTH,d2           ; depth
        moveq   #0,d3                   ; flags
        move.l  _BitMap,a0              ; friend bitmap
        move.l  _GfxBase,a6
        call    AllocBitMap
        move.l  d0,_tempbm
        beq     .nobitmap               ; close scr, fail
        lea     _temprp,a1
        move.l  _tempbm,(rp_BitMap,a1)
.no_temprp
        move.l  _RenderWidth,_ChunkyBPR
        move.l  _RenderWidth,d0
        mulu.w  _RenderHeight+2,d0
        add.l   #4096,d0                ; + 4K sanity buffer :)
        move.l  #$10001,d1              ; MEMF_CLEAR!MEMF_PUBLIC
        move.l  (4).w,a6
        call    AllocVec
        move.l  d0,_ChunkyPtr
        beq     .nomemory               ; FreeBitMap(tempbm), close scr, fail

        moveq   #1,d7                   ; return success
.exit
        move.l  d7,d0
        movem.l (sp)+,d1-a6
        rts

.nomemory
        move.l  _tempbm,a0
        move.l  _GfxBase,a6
        call    FreeBitMap
        clr.l   _tempbm
.nobitmap
        move.l  _Window,a0
        move.l  _IntuiBase,a6
        call    CloseWindow
        clr.l   _Window
        bra.b   .exit


.windowtags     dc.l    WA_Left,80
        dc.l    WA_Top,80

        dc.l    WA_IDCMP,IDCMP_CLOSEWINDOW
        dc.l    WA_Title,.title
        dc.l    WA_DragBar,1
        dc.l    WA_CloseGadget,1
        dc.l    WA_DepthGadget,1
        dc.l    WA_NoCareRefresh,1
        dc.l    WA_Activate,1
        dc.l    WA_RMBTrap,1

        dc.l    WA_ScreenTitle,.scrtitle
        dc.l    WA_AutoAdjust,1
        dc.l    WA_InnerWidth,SCR_WIDTH+2
        dc.l    WA_InnerHeight,SCR_HEIGHT+2

        dc.l    0

.title  dc.b    '#amycoders Gravity Particle compo',0
.scrtitle       dc.b    '#amycoders Gravity Particle compo entry by '
        dc.b    'Anynomous Coder <foo@bar.net>',0
        CNOP    0,4

        ELSE

openscreenetc
        movem.l d1-a6,-(sp)
        moveq   #0,d7

        ; First open screen with some tags
        ; Will get screen with default overscan.

        sub.l   a0,a0
        lea     (.screentags,pc),a1
        move.l  _IntuiBase,a6
        call    OpenScreenTagList
        move.l  d0,_Screen
        beq     .exit

        ; Cool, we have a screen! ;)
        ; Now we need to see whether we got native or
        ; GFXCard screen and choose between c2p and
        ; WriteChunkyPixels/WritePixelArray8.
        ; Here we go:

        move.l  _Screen,a0
        move.l  (sc_RastPort+rp_BitMap,a0),a0
        moveq   #BMA_FLAGS,d1           ; flags please
        move.l  _GfxBase,a6
        call    GetBitMapAttr

        and.w   #BMF_STANDARD,d0        ; custom highspeed c2p later...
;;      bne.b   .choose_custom

        ; Hmm, so apparently we have a GFXCard screen.
        ; Lets choose correct "c2p" for it.

        lea     _gfxcard_c2p_v39,a0
        move.l  (4).w,a6
        cmp.w   #40,(LIB_VERSION,a6)    ; V40 system?
        blo.b   .set_ptr                ; nopes, use WritePixelArray8
        lea     _gfxcard_c2p_v40,a0     ; yes, use WriteChunkyPixels
.set_ptr
        move.l  a0,_c2p_ptr
;;      bra     .after_cs
;;
;;.choose_custom
;;      ; Now we could check CPU and choose highspeed custom
;;      ; c2p optimized for it... In this example we have
;;      ; only some general c2p ;)
;;
;;      lea     _cpu_only_c2p,a0
;;      move.l  a0,_c2p_ptr
;;
;;.after_cs

        move.l  _Screen,a0
        lea     (sc_ViewPort,a0),a1
        move.l  a1,_ViewPort
        lea     (sc_RastPort,a0),a0
        move.l  a0,_RastPort
        move.l  (rp_BitMap,a0),a0
        move.l  a0,_BitMap
        move.l  (bm_Planes,a0),_Plane0

        move.l  _BitMap,a0
        moveq   #BMA_WIDTH,d1           ; width please
        move.l  _GfxBase,a6
        call    GetBitMapAttr
        move.l  d0,_Width
        lsr.l   #3,d0
        move.l  d0,_BytesPerRow

        move.l  _Screen,a0
        moveq   #0,d0
        move.w  (sc_Width,a0),d0
        move.l  d0,_RenderWidth

        move.l  _BitMap,a0
        moveq   #BMA_HEIGHT,d1          ; height please
        call    GetBitMapAttr
        move.l  d0,_Height

        move.l  _Screen,a0
        moveq   #0,d0
        move.w  (sc_Height,a0),d0
        move.l  d0,_RenderHeight

        move.l  _Height,d0
        move.l  _Width,d1
        mulu.w  d0,d1
        lsr.l   #3,d1
        move.l  d1,_PlaneSize           ; width*height/8

        cmp.l   #_gfxcard_c2p_v39,_c2p_ptr
        bne.b   .no_temprp

        ; temporary rastport (copy of rp with Layer set == NULL,
        ;temporary memory allocated for
        ;temprp->BitMap with Rows set == 1,
        ;temprp->BytesPerRow == (((width+15)>>4)<<1),
        ;and temporary memory allocated for
        ;temprp->BitMap->Planes[])

        move.l  _RastPort,a0
        lea     _temprp,a1
        move.l  #rp_SIZEOF,d0
        move.l  (4).w,a6
        call    CopyMem
        lea     _temprp,a1
        clr.l   (rp_Layer,a1)

        move.l  _RenderWidth,d0         ; width
        moveq   #1,d1                   ; height=1
        moveq   #SCR_DEPTH,d2           ; depth
        moveq   #0,d3                   ; flags
        move.l  _BitMap,a0              ; friend bitmap
        move.l  _GfxBase,a6
        call    AllocBitMap
        move.l  d0,_tempbm
        beq     .nobitmap               ; close scr, fail
        lea     _temprp,a1
        move.l  _tempbm,(rp_BitMap,a1)
.no_temprp
        move.l  _RenderWidth,_ChunkyBPR
        move.l  _RenderWidth,d0
        mulu.w  _RenderHeight+2,d0
        add.l   #4096,d0                ; + 4K sanity buffer :)
        move.l  #$10001,d1              ; MEMF_CLEAR!MEMF_PUBLIC
        move.l  (4).w,a6
        call    AllocVec
        move.l  d0,_ChunkyPtr
        beq     .nomemory               ; FreeBitMap(tempbm), close scr, fail

        moveq   #1,d7                   ; return success
.exit
        move.l  d7,d0
        movem.l (sp)+,d1-a6
        rts

.nomemory
        move.l  _tempbm,a0
        move.l  _GfxBase,a6
        call    FreeBitMap
        clr.l   _tempbm
.nobitmap
        move.l  _Screen,a0
        move.l  _IntuiBase,a6
        call    CloseScreen
        clr.l   _Screen
        bra.b   .exit


.screentags
        dc.l    SA_Width,SCR_WIDTH
        dc.l    SA_Height,SCR_HEIGHT
        dc.l    SA_Depth,SCR_DEPTH
        dc.l    SA_ShowTitle,0  ; no title please
        dc.l    SA_Quiet,-1     ; please don't draw anything, we'll do it
        dc.l    SA_Type,CUSTOMSCREEN
        ; see graphics/modeid.i for modes.
        ; no special monitor specified so will promote even
        ; with std Prefs/IControl promote.
        dc.l    SA_DisplayID,DEFAULT_MONITOR_ID!LORES_KEY
        dc.l    SA_Colors32,scr_colors
        dc.l    0               ; TAG_DONE

scr_colors
        dc.w    4,0             ; number of entries, 1st color to load

        dc.l    0,0,0                           ; black
        dc.l    $7FFFFFFF,$7FFFFFFF,$7FFFFFFF   ; grey (7f7f7f)
        dc.l    $FFFFFFFFF,0,0                  ; red (ff0000)
        dc.l    0,$FFFFFFFFF,0                  ; blue (00ff00)

        dc.w    0               ; num entries=0 -> end marker

        ENDC



        IFNE    WINDOW_MODE

_gfxcard_c2p_v39
        move.l  _RastPort,a0
        move.l  _RenderXstart,d0        ; xstart
        move.l  _RenderYstart,d1        ; ystart
        move.l  _RenderWidth,d2
        subq.l  #1,d2                   ; xstop
        add.l   d0,d2
        move.l  _RenderHeight,d3
        subq.l  #1,d3                   ; ystop
        add.l   d1,d3
        lea     _temprp,a1
        move.l  _ChunkyPtr,a2
        move.l  _GfxBase,a6
        jmp     (_LVOWritePixelArray8,a6)



        CNOP    0,8
_gfxcard_c2p_v40
        move.l  _RastPort,a0
        move.l  _RenderXstart,d0        ; xstart
        move.l  _RenderYstart,d1        ; ystart
        move.l  _RenderWidth,d2
        subq.l  #1,d2                   ; xstop
        add.l   d0,d2
        move.l  _RenderHeight,d3
        subq.l  #1,d3                   ; ystop
        add.l   d1,d3
        move.l  _ChunkyBPR,d4           ; bytesperrow (chunky)
        move.l  _ChunkyPtr,a2
        move.l  _GfxBase,a6
        jmp     (_LVOWriteChunkyPixels,a6)

        ELSE

        CNOP    0,8
_gfxcard_c2p_v39
        move.l  _RastPort,a0
        moveq   #0,d0                   ; xstart
        moveq   #0,d1                   ; ystart
        move.l  _RenderWidth,d2
        subq.l  #1,d2                   ; xstop
        move.l  _RenderHeight,d3
        subq.l  #1,d3                   ; ystop
        lea     _temprp,a1
        move.l  _ChunkyPtr,a2
        move.l  _GfxBase,a6
        jmp     (_LVOWritePixelArray8,a6)



        CNOP    0,8
_gfxcard_c2p_v40
        move.l  _RastPort,a0
        moveq   #0,d0                   ; xstart
        moveq   #0,d1                   ; ystart
        move.l  _RenderWidth,d2
        subq.l  #1,d2                   ; xstop
        move.l  _RenderHeight,d3
        subq.l  #1,d3                   ; ystop
        move.l  _ChunkyBPR,d4           ; bytesperrow (chunky)
        move.l  _ChunkyPtr,a2
        move.l  _GfxBase,a6
        jmp     (_LVOWriteChunkyPixels,a6)

        ENDC


        IFNE    WINDOW_MODE
closescreenetc
        movem.l d1-a6,-(sp)

        move.l  _GfxBase,a6
        call    WaitBlit

        move.l  _ChunkyPtr,a1
        move.l  (4).w,a6
        call    FreeVec
        clr.l   _ChunkyPtr

        move.l  _tempbm,a0
        move.l  _GfxBase,a6
        call    FreeBitMap
        clr.l   _tempbm

        move.l  (4).w,a6
.loop   move.l  _Window,a0
        move.l  (wd_UserPort,a0),a0
        call    GetMsg
        tst.l   d0
        beq.b   .ldone
        move.l  d0,a1
        call    ReplyMsg
        bra.b   .loop
.ldone

        move.l  _Window,a0
        move.l  _IntuiBase,a6
        call    CloseWindow
        clr.l   _Window

        movem.l (sp)+,d1-a6
        moveq   #1,d0
        rts


        ELSE
closescreenetc
        movem.l d1-a6,-(sp)

        move.l  _GfxBase,a6
        call    WaitBlit

        move.l  _ChunkyPtr,a1
        move.l  (4).w,a6
        call    FreeVec
        clr.l   _ChunkyPtr

        move.l  _tempbm,a0
        move.l  _GfxBase,a6
        call    FreeBitMap
        clr.l   _tempbm

        move.l  _Screen,a0
        move.l  _IntuiBase,a6
        call    CloseScreen
        clr.l   _Screen

        movem.l (sp)+,d1-a6
        moveq   #1,d0
        rts

        ENDC


        SECTION BSS,BSS

_GfxBase        ds.l    1
_IntuiBase      ds.l    1

_c2p_ptr        ds.l    1
        IFNE    WINDOW_MODE
_Window ds.l    1
_RenderXstart   ds.l    1
_RenderYstart   ds.l    1
        ELSE
_Screen ds.l    1
_ViewPort       ds.l    1
        ENDC
_RastPort       ds.l    1
_BitMap ds.l    1
_Plane0 ds.l    1
_BytesPerRow    ds.l    1
_Width  ds.l    1
_RenderWidth    ds.l    1
_Height ds.l    1
_RenderHeight   ds.l    1
_PlaneSize      ds.l    1
_ChunkyBPR      ds.l    1
_ChunkyPtr      ds.l    1
_tempbm ds.l    1
_temprp ds.b    rp_SIZEOF


        SECTION BSS,BSS

_ParticleArray  ds.w    NUM_PARTICLES*4


        SECTION CODE,CODE


; ******** YOUR CODE HERE ********

_gpart
;  IN: a0=pointer to chunky buffer (256x256x8)
;      a1=pointer to particle array {px1,py1,vx1,vy1...pxn,pyn,vxn,vyn}.w
* nAo/rAMjAM gravity compo contribution *
* 68 bytes version (good accuracy) *


gravity70       move.l  #1<<14-1,d7
                move.l  d7,d6
                
.loop           movem.w (a1),d0/d1/d2/d3
                asr.l   #4,d0
                asr.l   #4,d1
                move.w  d0,d4
                move.w  d1,d5
                muls.w  d4,d4
                muls.w  d5,d5
                asl.l   #8,d1
                move.b  d0,d1           
                cmp.l   d6,d4
                bgt.s   .skip_plot
                cmp.l   d6,d5
                bgt.s   .skip_plot
                move.b  d6,(a0,d1.w,128*256-128)
.skip_plot      asl.l   #8,d0
                add.w   d4,d5   
                beq.s   .skip_momentum
                divs.w  d5,d0
                divs.w  d5,d1
                sub.w   d0,d2
                sub.w   d1,d3
.skip_momentum  add.w   d2,(a1)+
                add.w   d3,(a1)+
                move.w  d2,(a1)+
                move.w  d3,(a1)+
.end            subq.w  #4,d7
                bpl.s   .loop
                rts




